//  Copyright (C) 2020 Éloïs SANCHEZ.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as
// published by the Free Software Foundation, either version 3 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

use serde::{Deserialize, Serialize};
use std::{
    net::{IpAddr, Ipv4Addr, Ipv6Addr},
    vec,
};

#[derive(Clone, Debug, Deserialize, Serialize)]
pub struct GvaConf {
    pub enabled: bool,
    #[serde(default = "ip4_default")]
    pub ip4: Ipv4Addr,
    pub ip6: Option<Ipv6Addr>,
    #[serde(default = "port_default")]
    pub port: u16,
    #[serde(default = "path_default")]
    pub path: String,
    #[serde(default = "subscriptions_path_default")]
    pub subscriptions_path: String,
    pub remote_host: Option<String>,
    pub remote_port: Option<u16>,
    pub remote_path: Option<String>,
    pub remote_subscriptions_path: Option<String>,
    pub remote_tls: Option<bool>,
    pub whitelist: Vec<IpAddr>,
}

const fn ip4_default() -> Ipv4Addr {
    Ipv4Addr::UNSPECIFIED
}

const fn ip6_default() -> Ipv6Addr {
    Ipv6Addr::UNSPECIFIED
}

fn path_default() -> String {
    "gva".to_owned()
}

const fn port_default() -> u16 {
    30_901
}

fn subscriptions_path_default() -> String {
    "gva-sub".to_owned()
}

impl Default for GvaConf {
    fn default() -> Self {
        GvaConf {
            enabled: false,
            ip4: ip4_default(),
            ip6: Some(ip6_default()),
            port: port_default(),
            path: path_default(),
            subscriptions_path: subscriptions_path_default(),
            remote_host: None,
            remote_port: None,
            remote_path: None,
            remote_subscriptions_path: None,
            remote_tls: None,
            whitelist: vec![
                IpAddr::V4(Ipv4Addr::LOCALHOST),
                IpAddr::V6(Ipv6Addr::LOCALHOST),
            ],
        }
    }
}

impl GvaConf {
    pub fn get_remote_port(&self) -> u16 {
        self.remote_port.unwrap_or(self.port)
    }
    pub fn get_remote_path(&self) -> String {
        self.remote_path
            .clone()
            .unwrap_or_else(|| self.path.clone())
    }
    pub fn get_remote_subscriptions_path(&self) -> String {
        self.remote_subscriptions_path
            .clone()
            .unwrap_or_else(|| self.subscriptions_path.clone())
    }
}
